from __future__ import annotations
from typing import Callable, List
from warnings import warn
import numpy as np
import scipy.sparse as sp
from pandas import Series as pd_Series
from pandas import DataFrame as pd_DataFrame
from polars import Series as pl_Series
from polars import DataFrame as pl_DataFrame
from tqdm import tqdm


def spyce_normalize(fun: Callable) -> Callable:
    def inner(mat: np.ndarray, **kwargs):
        shape = mat.shape
        norm_mat = fun(mat, **kwargs)
        if not len(norm_mat.shape) == len(shape) or any(
                [dim != shape[i_dim] for i_dim, dim in enumerate(norm_mat.shape)]):
            raise ValueError("Normalization changed shape of matrix which is not permitted.")
        else:
            return norm_mat
    return inner


def convert_normalization(normalization: Callable | str | None = "centered_sum") -> Callable:
    """
    Convert string or callable to spyce normalization function to make sure dimension is
    not changed. If passed as string, the function must be pre-implemented. If passed
    as callable, we add decorator @spyce_normalize.

    :param Callable | str | None normalization: Normalization function. If callable, make sure
        that the dimension is not changed. The framework will prevent you to perform a 
        normalization that changes the dimensionality of a matrix.
        If passed as a string, you can pass one of the following

            - `none`: No normalization is performed
            - `max_norm`: Largest value for each cell is set to 1
            - `sum_norm`: The kmer histogram for each cell sums up to 1
            - `center_norm`: Each sample is centered 
            - `centered_sum`: Unit sum histogram per cell and centering per sample
            - `centered_max`: Maximum value set to 1 per cell and centering per sample 
            - `centered_uniform_sum`: Centered unit sum normalization with unit variance 
            - `centered_uniform_max`: Maximum value per cell set to 1, centered, and unit variance
            
        Unless you know what you're doing, choose `centered_sum`
    :return Callable: normalization function
    """
    if normalization is None:
        normalization_fun = no_norm
    elif isinstance(normalization, str):
        normalization = normalization.lower()
        if normalization == "none" or normalization == "no_norm":
            normalization_fun = no_norm
        elif normalization == "max_norm":
            normalization_fun = max_norm
        elif normalization == "sum_norm":
            normalization_fun = sum_norm
        elif normalization == "center_norm":
            normalization_fun = center_norm
        elif normalization == "centered_sum":
            normalization_fun = centered_sum
        elif normalization == "centered_max":
            normalization_fun = centered_max
        elif normalization == "centered_uniform_sum":
            normalization_fun = centered_uniform_sum
        elif normalization == "centered_uniform_max":
            normalization_fun = centered_uniform_max
        else:
            raise ValueError("Normalization not understood. Choose between none | max_norm | "
                             "sum_norm | center_norm | centered_sum | centered_max | centered_uniform_sum | "
                             "centered_uniform_max.")
    elif callable(normalization):
        normalization_fun = spyce_normalize(normalization)
    else:
        raise ValueError("Normalization type not understood. Pass one of the following: none | str | callable.")
    return normalization_fun

def cscmat_division(
        mat: sp.lil_matrix | sp.csc_matrix,
        fun_str: str,
        axis: int | None = None,
        **kwargs
) -> sp.csc_matrix:
    """
    Helper function for sparse division
    """
    if fun_str.lower() == "sum":
        auxil_mat = sp.diags(1. / mat.sum(axis=axis).A.ravel())
    elif fun_str.lower() == "max":
        auxil_mat = sp.diags(1. / mat.max(axis=axis).A.ravel())
    else:
        raise ValueError("Function type not understood.")
    return sp.csc_matrix(auxil_mat.dot(mat))


@spyce_normalize
def no_norm(
    mat: np.ndarray | sp.lil_matrix | sp.csc_matrix,
    **kwargs
) -> np.ndarray | sp.lil_matrix | sp.csc_matrix:
    """
    Perform no normalization.

    :param np.ndarray | sp.lil_matrix | sp.csc_matrix mat: matrix to be normalized
    :param kwargs: Unused parameters.
    :return np.ndarray | sp.lil_matrix | sp.csc_matrix: Non-normalized matrix
    """
    return mat


@spyce_normalize
def max_norm(
    mat: np.ndarray | sp.lil_matrix | sp.csc_matrix,
    **kwargs
) -> np.ndarray | sp.lil_matrix | sp.csc_matrix:
    """
    Normalize such that the maximum value in each row is set to 1.

    :param np.ndarray | sp.lil_matrix | sp.csc_matrix mat: matrix to be normalized
    :param kwargs: Unused parameters.
    :return np.ndarray | sp.lil_matrix | sp.csc_matrix: Normalized matrix
    """
    if isinstance(mat, np.ndarray):
        return mat / mat.max(axis=1).reshape(-1, 1)
    else:
        return cscmat_division(mat, "max", axis=1)


@spyce_normalize
def sum_norm(
    mat: np.ndarray | sp.lil_matrix | sp.csc_matrix,
    **kwargs
) -> np.ndarray | sp.lil_matrix | sp.csc_matrix:
    """
    Normalize to unit sum per row.

    :param np.ndarray | sp.lil_matrix | sp.csc_matrix mat: matrix to be normalized
    :param kwargs: Unused parameters.
    :return np.ndarray | sp.lil_matrix | sp.csc_matrix: Normalized matrix
    """
    if isinstance(mat, np.ndarray):
        return mat / mat.sum(axis=1).reshape(-1, 1)
    else:
        return cscmat_division(mat, "sum", axis=1)

@spyce_normalize
def center_norm(
    mat: np.ndarray | sp.lil_matrix | sp.csc_matrix,
    use_median: bool = False,
    **kwargs
) -> np.ndarray | sp.lil_matrix | sp.csc_matrix:
    """
    Center matrix along columns.

    :param np.ndarray | sp.lil_matrix | sp.csc_matrix mat: matrix to be normalized
    :param bool use_median: If `True`, use the median for centering instead of mean.
    :param kwargs: Unused parameters.
    :return np.ndarray | sp.lil_matrix | sp.csc_matrix: Normalized matrix
    """
    if use_median and isinstance(mat, np.ndarray):
        norm_mat = mat - np.median(mat, axis=0)
    else:
        norm_mat = mat - mat.mean(axis=0).reshape(1, -1)
    # if not isinstance(mat, np.ndarray):
    #     norm_mat = sp.csc_matrix(norm_mat)
    return norm_mat

@spyce_normalize
def centered_sum(
    mat: np.ndarray | sp.lil_matrix | sp.csc_matrix,
    use_median: bool = False,
    **kwargs
) -> np.ndarray | sp.lil_matrix | sp.csc_matrix:
    """
    Centered unit sum normalization. Perform unit sum normalization before centering.

    :param np.ndarray | sp.lil_matrix | sp.csc_matrix mat: matrix to be normalized
    :param bool use_median: If `True`, use the median for centering instead of mean.
    :param kwargs: Unused parameters.
    :return np.ndarray | sp.lil_matrix | sp.csc_matrix: Normalized matrix
    """
    norm_mat = sum_norm(mat)
    return center_norm(norm_mat, use_median=use_median)


@spyce_normalize
def centered_max(
    mat: np.ndarray | sp.lil_matrix | sp.csc_matrix,
    use_median: bool = False,
    **kwargs
) -> np.ndarray | sp.lil_matrix | sp.csc_matrix:
    """
    Centered maximum value normalization. Perform maximum value normalization before centering.

    :param np.ndarray | sp.lil_matrix | sp.csc_matrix mat: matrix to be normalized
    :param bool use_median: If `True`, use the median for centering instead of mean.
    :param kwargs: Unused parameters.
    :return np.ndarray | sp.lil_matrix | sp.csc_matrix: Normalized matrix
    """
    norm_mat = max_norm(mat)
    return center_norm(norm_mat, use_median=use_median)


@spyce_normalize
def centered_uniform_sum(
    mat: np.ndarray | sp.lil_matrix | sp.csc_matrix,
    use_median: bool = False,
    use_std: bool = True,
    **kwargs
) -> np.ndarray | sp.lil_matrix | sp.csc_matrix:
    """
    Centered unit sum with unit spread. Perform centered unit sum normalization before transforming to unit spread.

    :param np.ndarray | sp.lil_matrix | sp.csc_matrix mat: matrix to be normalized
    :param bool use_median: If `True`, use the median for centering instead of mean.
    :param bool use_std: If `True`, normalize over standard deviation instead of variance.
    :param kwargs: Unused parameters.
    :return np.ndarray | sp.lil_matrix | sp.csc_matrix: Normalized matrix
    """
    norm_mat = centered_sum(mat, use_median=use_median)
    if isinstance(mat, np.ndarray):
        spread_fun = np.std if use_std else np.var
        spread = spread_fun(norm_mat, axis=0).reshape(1, -1)
        spread[spread == 0.] = 1.
        return norm_mat / spread
    else:
        warn("Can only unify spread for dense matrices safely. No division is performed.")
        return norm_mat

@spyce_normalize
def centered_uniform_max(
    mat: np.ndarray | sp.lil_matrix | sp.csc_matrix,
    use_median: bool = False,
    use_std: bool = True,
    **kwargs
) -> np.ndarray | sp.lil_matrix | sp.csc_matrix:
    """
    Centered unit sum with unit spread. Perform centered max value normalization before transforming to unit spread.

    :param np.ndarray | sp.lil_matrix | sp.csc_matrix mat: matrix to be normalized
    :param bool use_median: If `True`, use the median for centering instead of mean.
    :param bool use_std: If `True`, normalize over standard deviation instead of variance.
    :param kwargs: Unused parameters.
    :return np.ndarray | sp.lil_matrix | sp.csc_matrix: Normalized matrix
    """
    norm_mat = centered_max(mat, use_median=use_median)
    if isinstance(mat, np.ndarray):
        spread_fun = np.std if use_std else np.var
        spread = spread_fun(norm_mat, axis=0).reshape(1, -1)
        spread[spread == 0.] = 1.
        return norm_mat / spread
    else:
        warn("Can only unify spread for dense matrices safely. No division is performed.")
        return norm_mat

@spyce_normalize
def normalize_over_ord_vec(
    mat: np.ndarray | sp.lil_matrix | sp.csc_matrix | sp.csr_matrix | pd_DataFrame | pl_DataFrame,
    vec: List | np.ndarray | pl_Series | pd_Series,
    normalization: Callable | str | None = "centered_sum",
    **kwargs
) -> np.ndarray | sp.lil_matrix | sp.csc_matrix:
    """
    Normalize over arbitrary vector with ordinal values. The passed normalization is then applied per value in ordinal vector.

    :param np.ndarray | sp.lil_matrix | sp.csc_matrix mat: matrix to be normalized
    :param List | np.ndarray | pl_Series | pd_Series vec: Labels/categories used to group rows for normalization
    :param Callable | str | None normalization: Type of normalization performed.
    :param kwargs: Keyword arguments passed to the normalization function
    :return np.ndarray | sp.lil_matrix | sp.csc_matrix: Normalized matrix
    """
    if isinstance(vec, list):
        vec = np.array(vec).reshape(-1)
    if isinstance(vec, pl_Series) or isinstance(vec, pd_Series):
        vec = vec.to_numpy()
    if not isinstance(vec, np.ndarray):
        raise ValueError("Vector type not understood. Please pass `vec` as one of the following: "
                         "list | np.ndarray | polars Series | pandas Series.")
    vec = vec.reshape(-1)
    if vec.shape[0] != mat.shape[0]:
        raise ValueError("Vector length (%d) is unequal matrix length (%d). Note that vectors are flattened, "
                         "and the first matrix axis must be cell entries." % (vec.shape[0], mat.shape[0]))
    
    norm_fun = convert_normalization(normalization=normalization)
    cols, index = None, None
    if isinstance(mat, np.ndarray) or isinstance(mat, pd_DataFrame) or isinstance(mat, pl_DataFrame):
        norm_mat = np.zeros_like(mat)
        if isinstance(mat, pd_DataFrame):
            cols, index = mat.columns, mat.index
            mat = mat.to_numpy()
        elif isinstance(mat, pl_DataFrame):
            cols = mat.columns
            mat = mat.to_numpy()
    elif isinstance(mat, sp.lil_matrix):
        norm_mat = sp.lil_matrix(shape=mat.shape, dtype=mat.dtype)
    elif isinstance(mat, sp.csc_matrix):
        norm_mat = sp.csc_matrix(shape=mat.shape, dtype=mat.dtype)
    elif isinstance(mat, sp.csr_matrix):
        norm_mat = sp.csr_matrix(shape=mat.shape, dtype=mat.dtype)
    else:
        raise ValueError("Matrix type not understood. Please convert matrix to one of the following: "
                         "np.ndarray | scipy.sparse.lil_matrix | scipy.sparse.csc_matrix | scipy.sparse.csr_matrix.")
    
    for vec_val in np.unique(vec):
        mask = vec == vec_val 
        norm_mat[mask] = norm_fun(mat[mask], **kwargs)

    if cols is not None and index is not None:  # only set for pandas
        norm_mat = pd_DataFrame(norm_mat, index=index, columns=cols)
    elif cols is not None and index is None:  # only set for polars
        norm_mat = pl_DataFrame(norm_mat, schema=cols, orient="row")
    return norm_mat


@spyce_normalize
def harmony_correct(
    mat: np.ndarray | sp.lil_matrix | sp.csc_matrix,
    target_vec: np.ndarray | pl_Series | pd_Series | List,
    sigma: float = .3,
    max_iter: int = 20,
    theta: float | None = None,
    lamb: float | None = None,
    nclust: float | None = None,
    tau: float | None = 0.,
    block_size: float = 0.05, 
    max_iter_kmeans: int = 20,
    epsilon_cluster: float = 1e-5,
    epsilon_harmony: float = 1e-4, 
    **kwargs
) -> np.ndarray:
    """
    Batch or species correction using Harmony

    :param np.ndarray | sp.lil_matrix | sp.csc_matrix mat: matrix to be normalized
    :param List | np.ndarray | pl_Series | pd_Series target_vec: Labels/categories used for Harmony correction
    :param float sigma: Force by which soft-clusters are corrected. 
        The lower the value, the larger the correction.
    :param int max_iter: Maximum iterations
    :param float | None theta: Harmony theta parameter
    :param float | None lamb: Harmony lambda parameter
    :param int | None nclust: Number of clusters
    :param float tau: Harmony tau parameter
    :param float block_size: Harmony block size parameter
    :param int max_iter_kmeans: Maximum number of sklearn kmeans iterations
    :param float epsilon_cluster: Harmony epsilon cluster parameter
    :param float epsilon_harmony: Harmony epsilon harmony parameter
    :param kwargs: Unused parameters
    :return np.ndarray | sp.lil_matrix | sp.csc_matrix: Normalized matrix
    """
    import harmonypy
    if isinstance(target_vec, pl_Series):
        target_vec = target_vec.to_pandas()

    # assume that matrix is passed either dense or sparse
    if not isinstance(mat, np.ndarray):    
        mat = mat.A

    # run_harmony doesn't accept **kwargs. We need to set them specifically to allow 
    # other kwargs non related to harmony to be passed
    return harmonypy.run_harmony(
        mat,
        pd_DataFrame({"batch": target_vec}),
        "batch",
        sigma=sigma,
        max_iter_harmony=max_iter,
        theta=theta,
        lamb=lamb,
        nclust=nclust,
        tau=tau,
        block_size=block_size,
        max_iter_kmeans=max_iter_kmeans,
        epsilon_cluster=epsilon_cluster,
        epsilon_harmony=epsilon_harmony
    ).Z_corr.T


@spyce_normalize
def icp(
    mat: np.ndarray | sp.lil_matrix | sp.csc_matrix,
    target_vec:  np.ndarray | pl_Series | pd_Series | List,
    subset_frac: float = .6,
    tol: float = 1e-2,
    max_iter: int = 50,
    verbosity: int = 0,
    verbosity_indent: str = "",
    **kwargs
) -> np.ndarray | sp.lil_matrix | sp.csc_matrix:
    """
    Batch or species correction iterative closest point normalization as modified point-set registration transform.

    :param np.ndarray | sp.lil_matrix | sp.csc_matrix mat: matrix to be normalized
    :param List | np.ndarray | pl_Series | pd_Series target_vec: Labels/categories used for Harmony correction
    :param float subset_frac: Reduce run time by performing correction on subset
    :param float tol: Tolerance value. When update step lower than tolerance, stop ICP
    :param int max_iter: Maximum iterations
    :param int verbosity: Verbosity level
    :param str verbosity_indent: Prefix for output.
    :param kwargs: Unused parameters
    :return np.ndarray | sp.lil_matrix | sp.csc_matrix: Normalized matrix
    """
    from scipy.optimize import least_squares
    from sklearn.neighbors import KDTree
    def translation_cost(t_, mat1_, mat2_):
        return norm_fun((mat1_ - t_.reshape(1, -1)) - mat2_)
    
    if not 0 < subset_frac < 1.:
        raise ValueError("Subsampling fraction must be between 0 and 1")
    
    if isinstance(target_vec, pl_Series) or isinstance(target_vec, pd_Series):
        target_vec = target_vec.to_numpy()
    elif isinstance(target_vec, list):
        target_vec = np.array(target_vec)

    if isinstance(mat, sp.lil_matrix) or isinstance(mat, sp.csc_matrix):
        norm_fun = sp.linalg.norm
        trf_mat = sp.csc_matrix(shape=mat.shape)
    else:
        norm_fun = np.linalg.norm
        trf_mat = np.zeros_like(mat)
    src_mat = mat[~target_vec]
    dst_mat = mat[target_vec]
    # use point-set registration transform
    # find jointly correspondance between between data points of the two sets
    # and the corresponding transformation
    # intialize translational vector
    t_vec = src_mat.mean(axis=0).reshape(1, -1)
    difference = np.inf
    kdtree = KDTree(dst_mat)  
    epoch_iter = tqdm(
        range(max_iter), total=max_iter, desc="%sICP progress. Difference: inf."  % verbosity_indent
    ) if verbosity > 0 else range(max_iter) 
    for _ in epoch_iter:
        if verbosity > 0:
            epoch_iter.set_description(desc="%sICP progress. Difference: %.6f"  % (verbosity_indent, difference))

        subset_mat = src_mat[
            np.random.choice(src_mat.shape[0], replace=False, size=int(subset_frac * src_mat.shape[0]))
        ]
        t_mat = subset_mat - t_vec
        closest_match = np.zeros_like(t_mat)
        for i_m, m in enumerate(t_mat):
            closest_match[i_m] = dst_mat[kdtree.query(m.reshape(1, -1), return_distance=False)[0]]

        t_vec = least_squares(translation_cost, x0=t_vec.flatten(), args=(subset_mat, closest_match)).x
        t_vec = t_vec.reshape(1, -1)

        difference = norm_fun((subset_mat - t_vec) - closest_match)
        if difference <= tol:
            if verbosity > 0:
                print("%sFound registration, early stopping." % verbosity_indent)
            break

    trf_mat[~target_vec] = src_mat - t_vec
    trf_mat[target_vec] = dst_mat
    return trf_mat




from __future__ import annotations

import warnings
from typing import List, Tuple, Dict
from itertools import product

import numpy as np
import pandas as pd
import polars as pl
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
from matplotlib.lines import Line2D
import seaborn as sns

from spyce.kmerMatrix import KMerClass, KMerMatrix
from spyce.utils import nozero_kmer_to_idx
from spyce.constants import DNA_ALPHABET

def get_color(
        kmer_obj: KMerClass, 
        ck: str | pd.Series | pl.Series | np.ndarray | List[str], 
        cmin: int | float = 0.,
        cmax: int | float | None = None,
        cmap: str | None = None,
) -> np.ndarray:
    # convert polars series to pandas
    if isinstance(ck, pl.Series):
        ck = ck.to_pandas()
    # convert list to pandas
    if isinstance(ck, list):
        ck = pd.Series(ck)
    if isinstance(ck, pd.Series):
        # if string type and cmap is not passed, assume it's the color names
        if str(ck.dtype) in ["category", "str", "string"] and cmap is None:
            ck = ck.astype("category")
            return ck.to_numpy()
        
        # if cmap is passed, assume that ck is a categorical variable wrt which a color is
        # set using the cmap
        if str(ck.dtype) in ["category", "str", "string"] and cmap is not None:
            ck = ck.astype("category").cat.codes
        if "int" in str(ck.dtype) or "float" in str(ck.dtype):
            ck = ck.to_numpy()

    if isinstance(ck, np.ndarray) and cmap is None:
        # color matrix is set
        if len(ck.shape) == 3 or len(ck.shape) == 4:
            return ck
        # if one-dim numpy array, then it's a numerical value and cmap must be passed
        raise ValueError("If ck is a numpy array, you need to pass a cmap.")
    elif isinstance(ck, np.ndarray) and cmap is not None:
        c_vec = ck
    elif isinstance(ck, str):
        # fetch coloring
        c_vec = np.zeros(len(kmer_obj))
        ck = ck.lower()
        if ck == "species":
            if isinstance(kmer_obj, KMerMatrix):
                pass
            else:
                species_dict = {}
                idx = 0
                for i_mat, mat_values in enumerate(kmer_obj.kmer_mat_idc_df.iter_rows(named=False)):
                    species = kmer_obj.kmer_mat_list[i_mat].species
                    if species in species_dict:
                        species_idx = species_dict[species]
                    else:
                        species_idx = idx
                        idx += 1
                        species_dict[species] = species_idx
                    c_vec[int(mat_values[1]):int(mat_values[2])] = species_idx
        elif ck == "matrix":
            if isinstance(kmer_obj, KMerMatrix):
                pass
            else:
                for i_mat, mat_values in enumerate(kmer_obj.kmer_mat_idc_df.iter_rows(named=False)):
                    c_vec[int(mat_values[1]):int(mat_values[2])] = i_mat
        elif ck in kmer_obj.clustering.keys():
            c_vec = kmer_obj.clustering[ck]
        else:
            raise NotImplementedError("Accepting cell annotation keys is not yet implemented.")
    else:
        raise ValueError("Input type of ck not understood. Pass str | pd.Series | pl.Series | np.ndarray")
    
    norm = Normalize(vmin=cmin, vmax=cmax)
    if isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)
    return np.array([cmap(norm(val)) for val in c_vec])

def plot_scatter(
        X: np.ndarray,
        c_vec: np.ndarray | List,
        fig: plt.Figure | None = None,
        ax: plt.Axes | None = None,
        randomize: bool = True,
        title: str = "",
        s_vec: float | np.ndarray = 1.,
        x_type: str = "UMAP",
        dim: np.ndarray | None = None,
        return_scat: bool = False
) -> Tuple[plt.Figure, plt.Axes] | Tuple[plt.Figure, plt.Axes, List]:
    """
    General function for creating a scatter plot using matplotlib. We recommend using only
    one of the wrappers provided.

    :param np.ndarray X: data (if size `#samples x #features`)
    :param np.ndarray | List c_vec: Color vector with one entry per data point
    :param plt.Figure | None fig: Matplotlib figure. Pass `None` to create a new figure.  
    :param plt.Axes | None ax: Matplotlib axes. Pass `None` to create new axes.
    :param bool randomize: If set to `True`, randomize order in which data points are plotted. 
    :param str title: Axes title
    :param float | np.ndarray s_vec: Size of plotted circles, can be a single float or a vector
        per data entry in `X`.
    :param str x_type: Type of representation that is used for labelling the x and y axes.
    :param np.ndarray | None dim: Dimension names for obtaining `features` in `X`.
    :param bool return_scat: Return the Matplotlib scatter object 
    :return Tuple[plt.Figure, plt.Axes] | Tuple[plt.Figure, plt.Axes, List]: If `return_scat=False`
        return only figure and axes, otherwise return figure, axes, and scatter object.
    """
    if ax is None or fig is None:
        fig = plt.figure()
        
    idc = np.arange(X.shape[0])
    if randomize:
        # shuffle data points
        np.random.shuffle(idc)

    if X.shape[1] < 3:
        # 1D and 2D plotting
        if ax is None:
            ax = fig.add_subplot(111)
        x_val = X[:, 0]
        y_val = np.arange(X.shape[0]) if X.shape[1] == 1 else X[:, 1]
        scat = ax.scatter(
            x_val[idc],
            y_val[idc],
            marker=".",
            s=s_vec,
            c=c_vec[idc],
        )
        if dim is None:
            dim = np.array([1, 2])
        else: 
            dim = dim + 1
        ax.set_xlabel("%s%d" % (x_type, dim[0]))
        ax.set_ylabel("%s%d" % (x_type, dim[1]))
        ax.set_xticks([])
        ax.set_yticks([])

    else:
        # 3D plotting
        if ax is None:
            ax = fig.add_subplot(111, projection="3d")
        else:
            warnings.warn("Cannot perform 3D projection when axis is passed.")
            return None, None, None
        scat = ax.scatter(
            X[idc, 0],
            X[idc, 1],
            X[idc, 2],
            marker=".",
            s=s_vec,
            c=c_vec[idc],
        )
        if dim is None:
            dim = np.array([1, 2, 3])
        else: 
            dim = dim + 1
        ax.set_xlabel("%s%d" % (x_type, dim[0]))
        ax.set_ylabel("%s%d" % (x_type, dim[1]))
        ax.set_zlabel("%s%d" % (x_type, dim[2]))
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_zticks([])

    ax.set_title(title)
    if return_scat:
        return fig, ax, scat
    else:
        return fig, ax


def plot_dr(
    kmer_obj: KMerClass,
    ck: str | pd.Series | pl.Series,
    dr_key: str = "pca",
    dim: np.ndarray = np.array([0, 1]),
    fig: plt.Figure | None = None,
    ax: plt.Axes | None = None,
    randomize: bool = True,
    title: str = "",
    cmin: int | float = 0.,
    cmax: int | float | None = None,
    cmap: str = "tab20",
    s_vec: float | np.ndarray = 1.,
    return_scat: bool = False
) -> Tuple[plt.Figure, plt.Axes] | Tuple[plt.Figure, plt.Axes, List]:
    """
    Plot data in KMer data object and color with respect to key

    :param KMerClass kmer_obj: KMer data object, either of type KMerMatrix or KMerCollection
    :param str ck: Color key which defines the coloring of the data points. Can be set to pne of the following:
        
        - `species`: Color w.r.t. species. Only useful if the KMer data object is of type KMerCollection
        - `matrix`: Color w.r.t. each matrix. Only useful if the KMer data object is of type KMerCollection
        - `your_cluster_key`: Cluster w.r.t. a clustering result saved under `your_cluster_key`. You must run a
            clustering algorithm before, which is saved under this name
        - `your_cell_annotation_col`: Cluster w.r.t. a column in the cell annotation table `your_cell_annotation_col`.
            This is only useful if a cell annotation table was set
        
    :param str dr_key: Key name for the dimensionality reduction to be used. This is a key in the `kmer_obj.dr` dict.
    :param np.ndarray dim: Dimensions to be used for plotting.
    :param plt.Figure | None fig: Figure which is used for plotting
    :param plt.Axes | None ax: Axis used for plotting
    :param bool randomize: If true, randomize data values for plotting
    :param str title: Title of axis
    :param int | float cmin: Minimum color value which is used for normalizing
    :param int | float | None cmax: Maximum color value. If not passed, use the maximum value found color key
    :param str cmap: Name of color map. See https://matplotlib.org/stable/users/explain/colors/colormaps.html
    :param float | np.ndarray s_vec: Size of plotted circles, can be a single float or a vector
        per data entry in `X`.
    :param bool return_scat: Return the Matplotlib scatter object 
    :return Tuple[plt.Figure, plt.Axes] | Tuple[plt.Figure, plt.Axes, List]: If `return_scat=False`
        return only figure and axes, otherwise return figure, axes, and scatter object.
    """
    if kmer_obj.dr is None:
        raise ValueError("Plotting requires a calculating a dimensionality reduction first.")

    if dr_key not in kmer_obj.dr:
        raise ValueError("Provided dimensionality reduction key doesn't exist.")
    
    if not 1 <= len(dim) <= 3:
        raise ValueError("Plotting requires a dimensionality reduction into a 1, 2, or 3-D space.")
    
    dim = np.array(dim)
    # get data
    x_dr = kmer_obj.dr[dr_key][:, dim]
    # get color
    c_vec = get_color(kmer_obj=kmer_obj, ck=ck, cmin=cmin, cmax=cmax, cmap=cmap)
    # plot
    return_vals = plot_scatter(
        X=x_dr,
        c_vec=c_vec,
        fig=fig,
        ax=ax,
        randomize=randomize,
        title=title,
        s_vec=s_vec,
        dim=dim,
        x_type=dr_key,
        return_scat=return_scat
    )
    return return_vals


def plot_umap(
        kmer_obj: KMerClass,
        ck: str,
        umap_key: str = "umap",
        fig: plt.Figure | None = None,
        ax: plt.Axes | None = None,
        randomize: bool = True,
        title: str = "",
        cmin: int | float = 0.,
        cmax: int | float | None = None,
        cmap: str = "tab20",
        s_vec: float | np.ndarray = 1.,
        return_scat: bool = False
) -> Tuple[plt.Figure, plt.Axes] | Tuple[plt.Figure, plt.Axes, List]:
    """
    Plot data in KMer data object and color with respect to key

    :param KMerClass kmer_obj: KMer data object, either of type KMerMatrix or KMerCollection
    :param str ck: Color key which defines the coloring of the data points. Can be set to pne of the following:
        
        - `species`: Color w.r.t. species. Only useful if the KMer data object is of type KMerCollection
        - `matrix`: Color w.r.t. each matrix. Only useful if the KMer data object is of type KMerCollection
        - `your_cluster_key`: Cluster w.r.t. a clustering result saved under `your_cluster_key`. You must run a
            clustering algorithm before, which is saved under this name
        - `your_cell_annotation_col`: Cluster w.r.t. a column in the cell annotation table `your_cell_annotation_col`.
            This is only useful if a cell annotation table was set
    
    :param str umap_key: Key name for the UMAP embedding to be used. This is a key in the `kmer_obj.x_umap` dict.
    :param plt.Figure | None fig: Figure which is used for plotting
    :param plt.Axes | None ax: Axis used for plotting
    :param bool randomize: If true, randomize data values for plotting
    :param str title: Title of axis
    :param int | float cmin: Minimum color value which is used for normalizing
    :param int | float | None cmax: Maximum color value. If not passed, use the maximum value found color key
    :param str cmap: Name of color map. See https://matplotlib.org/stable/users/explain/colors/colormaps.html
    :param float | np.ndarray s_vec: Size of plotted circles, can be a single float or a vector
        per data entry in `X`.
    :param bool return_scat: Return the Matplotlib scatter object 
    :return Tuple[plt.Figure, plt.Axes] | Tuple[plt.Figure, plt.Axes, List]: If `return_scat=False`
        return only figure and axes, otherwise return figure, axes, and scatter object.
    """
    if kmer_obj.x_umap is None:
        raise ValueError("Plotting requires a calculating the UMAP first.")

    if umap_key not in kmer_obj.x_umap:
        raise ValueError("Provided UMAP key doesn't exist.")
    
    if not 0 < kmer_obj.x_umap[umap_key].shape[1] <= 3:
        raise ValueError("Plotting requires an embedding into 1, 2, or 3-D manifold.")

    x_umap = kmer_obj.x_umap[umap_key]
    # get color
    c_vec = get_color(kmer_obj=kmer_obj, ck=ck, cmin=cmin, cmax=cmax, cmap=cmap)
    # plot
    return_vals = plot_scatter(
        X=x_umap,
        c_vec=c_vec,
        fig=fig,
        ax=ax,
        randomize=randomize,
        title=title,
        s_vec=s_vec,
        x_type=umap_key,
        return_scat=return_scat
    )
    return return_vals
    
    
def plot_clustermap(
        kmer_obj: KMerClass,
        row_colors: pd.Series | List[pd.Series] | None = None,
        col_colors: pd.Series | List[pd.Series] | None = None,
        cmin: float | int | None = None,
        cmax: float | int | None = None,
        col_cluster: bool = False,
        row_cluster: bool = True,
        clustermap_kw: Dict = {}
) -> sns.matrix.ClusterGrid:
    """
    Plot k-mer histogram as clustered heatmap using seaborn.

    :param KMerClass kmer_obj: KMer object whose cell/nucleus specific kmer histogram
        is used for plotting
    :param pd.Series | List[pd.Series] | None row_colors: Use these colors for idenifying rows.
    :param pd.Series | List[pd.Series] | None col_colors: Use these colors for idenifying columns.
    :param float | int | None cmin: Minimum color value
    :param float | int | None cmax: Maximum color value
    :param bool col_cluster: If set to true, cluster columns and visualize with dendrogram
    :param bool row_cluster: If set to true, cluster row and visualize with dendrogram
    :param dict clustermap_kw: Parameters that are passed to the cluster map function provided by Seaborn.
    :return sns.matrix.ClusterGrid: Seaborn cluster grid.
    """
    kmer_hist_hm = pd.DataFrame(kmer_obj.kmer_hist)
    kmer_hist_hm.columns = kmer_obj.get_kmer_column_names()

    cg = sns.clustermap(
        kmer_hist_hm,
        col_cluster=col_cluster,
        row_cluster=row_cluster,
        row_colors=row_colors,
        col_colors=col_colors,
        vmax=cmax,
        vmin=cmin,
        **clustermap_kw
    )
    return cg


from __future__ import annotations
import os
from typing import List, Dict, Callable, Tuple
from itertools import product
import warnings
import numpy as np
import anndata as ad
import scipy.sparse as sp
import scipy.stats as st
from sklearn.decomposition import PCA
import pandas as pd
import polars as pl
from gimmemotifs.motif import read_motifs
from gimmemotifs.config import DIRECT_NAME, INDIRECT_NAME
from Bio.Seq import Seq
from tqdm import tqdm

from spyce.kmerMatrix import KMerClass, KMerMatrix
from spyce.constants import DNA_ALPHABET
from spyce.utils import fetch_async_result, seq_to_int, nozero_kmer_to_idx
from spyce.normalize import convert_normalization, harmony_correct
import multiprocessing


def chi2_pwm_convolve_diff(pwm1: np.ndarray, pwm2: np.ndarray, thresh: float = .95, min_conv: int = 4) -> float:
    """
    Computes similarity between two convoluting TFBS motifs given PWM using chi-squared significant test.
    Determines the number of positions that are significantly similar or dissimilar, dependening
    on the set threshold. Metric taken and adapted from https://doi.org/10.1186/1471-2105-6-237

    :param np.ndarray pwm1: PWM of first motif
    :param np.ndarray pwm2: PWM of second motif
    :param float thresh: p-value threshold used for defining whether the pwm at a given position is 
        sufficiently similar. For example, if the threshold is set to 0.95, it measures significant
        similarity. The distance is therefore the fraction of positions that are not significantly similar.
        If the threshold is set to 0.05, the measure represent significant dissimilarity. The
        distance is the fraction of positions that are significantly different.
    :param int min_conv: Minimum overlap between the motifs.
    :return float: Difference between motif1 and 2 based on their PWMs 
    """
    # get shorter pwm
    shorter_pwm, longer_pwm = (pwm1, pwm2) if pwm1.shape[0] < pwm2.shape[0] else (pwm2, pwm1)
    min_size = shorter_pwm.shape[0]
    max_size = longer_pwm.shape[0]
    # make sure probabilities sum up to one
    pwm1 /= pwm1.sum(axis=1).reshape(-1, 1)
    pwm2 /= pwm2.sum(axis=1).reshape(-1, 1)
    dist_list = []
    # convolute the two pwms
    for lag in range(min_conv, pwm1.shape[0] + pwm2.shape[0] - (min_conv + 1)):
        # get end of longer pwm
        cropped_longer = longer_pwm[-np.minimum(lag, max_size)
                                    :-(lag - min_size) if (lag - min_size) > 0 else None]
        # get beginning of shorter pwm
        cropped_shorter = shorter_pwm[
            np.maximum(0, lag - longer_pwm.shape[0])
            :np.minimum(lag, min_size)]
        
        # chi squared significance test whether observed frequencies match given distribution.
        # note that chisquare is asymmetric. As the definition of observation and distribution is
        # arbitrary, we perform both sides
        chisq_1 = np.nan_to_num(st.chisquare(cropped_shorter, cropped_longer, axis=1).pvalue, nan=1.)
        chisq_2 = np.nan_to_num(st.chisquare(cropped_longer, cropped_shorter, axis=1).pvalue, nan=1.)
        
        # take the maximum probability out of the two chi squared tests.
        # this indicates probability that the pwms are different at a given position.
        # look how many are larger than threshold on average given the size of the motif.
        # if threshold is 0.95, this represents the probability that they are significantly similar
        # if threshold is 0.05, this respresents probability that they are significantly dissimilar
        dist_list.append((np.maximum(chisq_1, chisq_2) >= thresh).mean())    
    # the larger max(dist_list), the more similar are motifs. Average of p-values ranges [0, 1].
    # as dissimilarity metrc, we compute 1 - max(dist_list
    return 1. - max(dist_list)


def chi2_pwm_diff(pwm1: np.ndarray, pwm2: np.ndarray, thresh: float = .95) -> float:
    """
    Computes similarity between two TFBS motifs given PWM using chi-squared significant test.
    Determines the number of positions that are significantly similar or dissimilar, dependening
    on the set threshold. Metric taken and adapted from https://doi.org/10.1186/1471-2105-6-237

    :param np.ndarray pwm1: PWM of first motif
    :param np.ndarray pwm2: PWM of second motif
    :param float thresh: p-value threshold used for defining whether the pwm at a given position is 
        sufficiently similar. For example, if the threshold is set to 0.95, it measures significant
        similarity. The distance is therefore the fraction of positions that are not significantly similar.
        If the threshold is set to 0.05, the measure represent significant dissimilarity. The
        distance is the fraction of positions that are significantly different.
    :return float: Difference between motif1 and 2 based on their PWMs 
    """
    # get shorter pwm
    shorter_pwm, longer_pwm = (pwm1, pwm2) if pwm1.shape[0] < pwm2.shape[0] else (pwm2, pwm1)
    # make sure probabilities sum up to one
    shorter_pwm /= shorter_pwm.sum(axis=1).reshape(-1, 1)
    longer_pwm /= longer_pwm.sum(axis=1).reshape(-1, 1)
    dist_list = []
    # move shorter motif along larger motif and compare significant differences
    for lag in range(0, longer_pwm.shape[0] - shorter_pwm.shape[0] + 1):
        # crop larger motif to same size
        cropped_pwm = longer_pwm[lag:lag + shorter_pwm.shape[0]]
        # chi squared significance test whether observed frequencies match given distribution.
        # note that chisquare is asymmetric. As the definition of observation and distribution is
        # arbitrary, we perform both sides
        chisq_1 = np.nan_to_num(st.chisquare(cropped_pwm, shorter_pwm, axis=1).pvalue, nan=1.)
        chisq_2 = np.nan_to_num(st.chisquare(shorter_pwm, cropped_pwm, axis=1).pvalue, nan=1.)
        # take the maximum probability out of the two chi squared tests.
        # this indicates probability that the pwms are different at a given position.
        # look how many are larger than threshold on average given the size of the motif.
        # if threshold is 0.95, this represents the probability that they are significantly similar
        # if threshold is 0.05, this respresents probability that they are significantly dissimilar
        dist_list.append((np.maximum(chisq_1, chisq_2) >= thresh).mean())    
    # the larger max(dist_list), the more similar are motifs. Average of p-values ranges [0, 1].
    # as dissimilarity metrc, we compute 1 - max(dist_list)
    return 1. - max(dist_list)


def motif_kmer_score(
        ppm: np.ndarray,
        k: int = 6,
        alphabet: List[str] = DNA_ALPHABET,
        return_dense: bool = True,
        rm_zeros: bool = False,
        use_thresh: bool = False,
        min_prob: float | None = None,
        collapse_fun: Callable = np.maximum,
        consensus_thresh: float = .0,
        equalize_counter: bool = True
) -> np.ndarray | sp.lil_matrix:
    """
    Helper function that returns kmer probability scores for a given transcription factor when provided with
    a position probability matrix. This is called during parallel processing.

    :param np.ndarray ppm: Position probability matrix
    :param int k: K-mer size
    :param List[str] alphabet: Letters in k-mer alphabet
    :param bool return_dense: If set, return dense numpy array instead of sparse matrix
    :param bool rm_zeros: If set, remove reverse complements if `equalize_counter=True`. For example,
        assuming `k=3`, `AAT` and `ATT` (the reverse complement) would both contain the same value.
        They can be removde to save memory. Note, however, that in many use cases, it is easier to keep
        them.
    :param bool use_thresh: If set, probability values for a k-mer to be present in motif must be larger
        than `min_prob`, or are set to zero.
    :param float | None min_prob: If set and `use_thresh`, use this value as minimum probability for k-mer
        being present in motif, otherwise set to zero. If `None`, and `use_thresh=True` set
        `min_prob = k * .25**k`.
    :param Callable collapse_fun: Probability of `k` tuples are multiplied per position. `collapse_fun`
        defines how the position-specific information is combined with the position-oblivious k-mer histogram
        representation. For example, you can add them all together (by passing for example `np.add`, or you
        keep only the maximum value (by passing for example `np.maximum`). The passed callable must accept
        to parameters, the first is the k-mer histogram computed up to the n-th position, and the probility
        of finding the `k`-tuple at positon `n+1`. Default is `np.add`. 
    :param float consensus_thresh: Minimum probability for a nucleotide to be present in order to be included into k-mer
        probability, otherwise set to zero.
    :param bool equalize_counter: If set to `True`, treat forward and reverse complement kmer as the same kmer and add
        probability to both entries. If `rm_zeros=True`, add only to first alphabetical k-mer. That
        means values for `GCGT` are added to the histogram values for `ACGC` if `rm_zeros=True`.
    :return np.ndarray | sp.lil_matrix: Numpy array or sparse matrix depending on whether `return_dense` is
        set to true.
    """
    motif_kmer_total = np.zeros((1, len(alphabet)**k))
    if min_prob is None:
        min_prob = k * .25**k
    for i in range(0, ppm.shape[0] - k + 1):
        to_idx = np.minimum(k + i,  ppm.shape[0])
        # get all kmer combinations for these positions as indices in ppm
        col_idc = np.array(list(product(np.arange(len(alphabet)), repeat=(to_idx - i))))
        row_idc = np.tile(np.arange(i, to_idx), reps=col_idc.shape[0])
        kmer_ppm = ppm[row_idc.flatten(), col_idc.flatten()].reshape(-1, (to_idx - i))
        kmer_ppm[kmer_ppm < consensus_thresh] = 0.

        # multiply position probabilities per kmer
        motif_kmer_total[0, np.arange(len(alphabet)**(to_idx - i))] = collapse_fun(
            motif_kmer_total[0, np.arange(len(alphabet)**(to_idx - i))],
            np.prod(kmer_ppm, axis=1)
        )
    
    if equalize_counter:
        # account for the fact that the annotation could also occur for the reversed strand
        equalized_kmer_counter = {}
        # sort alphabetically
        for kmer in product(alphabet, repeat=k):
            kmer = "".join(kmer)
            rev_kmer = str(Seq(kmer).reverse_complement())
            # make sure order is alphabetical
            if rm_zeros:
                save_kmer = kmer if kmer < rev_kmer else rev_kmer
                idx = seq_to_int(kmer, alphabet=alphabet)
                equalized_kmer_counter[save_kmer] = motif_kmer_total[0, idx] + equalized_kmer_counter.get(rev_kmer, 0)
            else:
                if kmer in equalized_kmer_counter or rev_kmer in equalized_kmer_counter:
                    continue
                idx = seq_to_int(kmer, alphabet=alphabet)
                rev_idx = seq_to_int(rev_kmer, alphabet=alphabet)
                equalized_kmer_counter[kmer] = motif_kmer_total[0, idx] + motif_kmer_total[0, rev_idx]
                equalized_kmer_counter[rev_kmer] = motif_kmer_total[0, idx] + motif_kmer_total[0, rev_idx]
    else:
        equalized_kmer_counter = dict(zip(
            ["".join(kmer) for kmer in product(alphabet, repeat=k)],
            motif_kmer_total.reshape(-1)
        ))

    kmer_list, n_kmer = zip(*sorted(equalized_kmer_counter.items(), key=lambda _: _[0]))
    # convert present kmers to unique numerical value
    if rm_zeros:
        kmer_idx_dict = nozero_kmer_to_idx(alphabet=alphabet, k=k)
        kmer_idx = [kmer_idx_dict[kmer] for kmer in kmer_list]
    else:
        kmer_idx = [seq_to_int(kmer, alphabet) for kmer in kmer_list]
    # set kmer histogram values
    motif_kmer = np.zeros(max(kmer_idx_dict.values()) + 1 if rm_zeros else len(alphabet)**k, dtype="float")
    motif_kmer[kmer_idx] = n_kmer

    if equalize_counter:
        motif_kmer /= 2.

    if use_thresh:
        motif_kmer[motif_kmer < min_prob] = 0.

    motif_kmer = motif_kmer.reshape(1, -1)
    return sp.coo_matrix(motif_kmer).reshape(1, -1) if not return_dense else motif_kmer


def motifs_to_kmer(
        k: int = 6,
        alphabet: List[str] = DNA_ALPHABET,
        db_name: str = "CIS-BP",
        return_dense: bool = False,
        include_indirect: bool = True,
        n_jobs: int = 1,
        rm_zeros: bool = True,
        use_thresh: bool = False,
        min_prob: float | None = None,
        collapse_fun: Callable = np.add,
        normalization: Callable | str | None = "none",
        consensus_thresh: float = .0,
        equalize_counter: bool = True,
        pca_transform: PCA | None = None,
        verbosity: int = 0,
        verbosity_indent: str = "",
) -> KMerMatrix:
    """
    Convert transcription factor binding motifs from a data base to a k-mer matrix which is 
    returned as and AnnData object (rather than a KMer object).

    :param int k: k-mer size
    :param List[str] alphabet: k-mer alphabet
    :param str db_name: transcription factor binding site motif data base name. Choose a motif supported
        by the gimmemotifs library, which are CIS-BP | ENCODE | factorbook | HOCOMOCOv11_HUMAN |
        HOCOMOCOv11_MOUSE | HOMER | IMAGE | JASPAR[2018 | 2020][_vertebrates | _plants | _insects | _fungi 
        _nematodes | _urochordata]
    :param bool return_dense: If set, use a dense numpy array instead a sparse matrix
    :param bool include_indirect: also include motifs for which there is only computational support
    :param int n_jobs: Number of processes used
    :param bool rm_zeros: If set, remove reverse complements if `equalize_counter=True`. For example,
        assuming `k=3`, `AAT` and `ATT` (the reverse complement) would both contain the same value.
        They can be removde to save memory. Note, however, that in many use cases, it is easier to keep
        them.
    :param bool use_thresh: If set, probability values for a k-mer to be present in motif must be larger
        than `min_prob`, or are set to zero.
    :param float | None min_prob: If set and `use_thresh`, use this value as minimum probability for k-mer
        being present in motif, otherwise set to zero. If `None`, and `use_thresh=True` set
        `min_prob = k * .25**k`.
    :param Callable collapse_fun: Probability of `k` tuples are multiplied per position. `collapse_fun`
        defines how the position-specific information is combined with the position-oblivious k-mer histogram
        representation. For example, you can add them all together (by passing for example `np.add`, or you
        keep only the maximum value (by passing for example `np.maximum`). The passed callable must accept
        to parameters, the first is the k-mer histogram computed up to the n-th position, and the probility
        of finding the `k`-tuple at positon `n+1`. Default is `np.add`. 
    :param str | Callable | None normalization: normalization function. Can be either a user defined callable;
        or one of the following:

        - None or `none`: no normalization
        - `max_norm`: normalize every cell by its maximum value
        - `sum_norm`: normalize every cell by its sum
        - `centered_sum`: normalize every cell by its sum and then center every histogram value
        - `centered_max`: normalize every cell by its maximum value and then center every histogram value
        - `centered_uniform_sum`: normalize every cell by its sum and then center every histogram value with
            uniform variance
        - `centered_uniform_max`: normalize every cell by its maximum value and then center every histogram value with
            uniform variance    

    :param float consensus_thresh: Minimum probability for a nucleotide to be present in order to be included into k-mer
        probability, otherwise set to zero.
    :param bool equalize_counter: If set to `True`, treat forward and reverse complement kmer as the same kmer and add
        probability to both entries. If `rm_zeros=True`, add only to first alphabetical k-mer. That
        means values for `GCGT` are added to the histogram values for `ACGC` if `rm_zeros=True`.
    :param PCA | None pca_transform: sklearn PCA object. If passed, perform dimensionality reduction on pre-trained PCA
    :param int verbosity: verbosity level
    :param str verbosity_indent: indentation string added before printout
    :return KMerMatrix: k-mer histrogram matrix as KMerMatrix object
    """
    with multiprocessing.get_context("spawn").Pool(n_jobs) as pool:
        if verbosity > 0:
            print("Fetch motifs from %s" % db_name)
        tf_motifs = read_motifs(db_name)
        motif_jobs = []
        if verbosity > 0:
            print("Prepare motif k-mers.")

        tf_iter = tqdm(
            tf_motifs,
            total=len(tf_motifs),
            desc="Compute TF probability"
        ) if verbosity > 0 else tf_motifs

        motif_name_dict = {}
        motif_results = []
        for motif in tf_iter:
            # get ppm 
            ppm = motif.ppm
            # get motif name
            factor_name = motif.factors[DIRECT_NAME]
            indirect_name = motif.factors[INDIRECT_NAME]
            if len(factor_name) == 0 and len(indirect_name) == 0:
                factor_name = motif.id
            elif len(factor_name) > 0:
                factor_name = factor_name[0] 
            else:
                if include_indirect:
                    factor_name = indirect_name[0]
                else:
                    continue
            motif_name_dict[motif.id] = factor_name
            if n_jobs > 1:
                # create job
                motif_jobs.append((motif.id, pool.apply_async(
                    motif_kmer_score, args=(
                        ppm, 
                        k, 
                        alphabet,
                        return_dense,
                        rm_zeros,
                        use_thresh,
                        min_prob,
                        collapse_fun,
                        consensus_thresh,
                        equalize_counter
                    ))))
            else:
                motif_results.append((motif.id, motif_kmer_score(
                            ppm, 
                            k, 
                            alphabet,
                            return_dense,
                            rm_zeros,
                            use_thresh,
                            min_prob,
                            collapse_fun,
                            consensus_thresh,
                            equalize_counter
                        )))

                
        if n_jobs > 1:
            # fetch completed jobs   
            pbar = tqdm(
                total=len(tf_motifs),
                desc="%sFetch motif kmers" % verbosity_indent
            ) if verbosity > 0 else None
            motif_results = fetch_async_result(motif_jobs, process_bar=pbar)
        stack_lib = np if return_dense else sp
        motif_id_list = []
        kmer_list = []
        for motif_id, kmer in sorted(motif_results, key=lambda x: x[0]):
            motif_id_list.append(motif_id)
            kmer_list.append(kmer)
        # create var name list, which are the factor IDs
        if rm_zeros:
            kmer_seq_list = list(sorted(nozero_kmer_to_idx(alphabet=alphabet, k=k).keys()))
        else:
            kmer_seq_list = product(alphabet, repeat=k)
        kmer_tuple_list = []
        for kmer_seq in kmer_seq_list:
            kmer_tuple_list.append((seq_to_int(kmer_seq, alphabet=alphabet), "".join(kmer_seq)))
        
        # create anndata object
        motif_kmer_mat = stack_lib.vstack(kmer_list)
        motif_kmer = KMerMatrix(
            kmer_hist=motif_kmer_mat,
            barcodes=pd.Series(motif_id_list),
            k=k,
            normalization=normalization
        )
        if pca_transform is not None:
            motif_kmer.dr = {}
            motif_kmer.dr["ext_pca"] = pca_transform.transform(motif_kmer.kmer_hist)
        return motif_kmer


def data_set_motif_score(
        kmer_obj: KMerClass,
        motif_kmer: KMerClass,
        use_kmer_dr: str | None = None,
        use_motif_dr: str | None = None,
        norm: Callable | str | None = None,
        norm_vec: pd.Series | pl.Series | np.ndarray | List | None = None,
        harmony_vec: pd.Series | pl.Series | np.ndarray | List | None = None,
        collapse_fun: str | Callable = "sum",
        verbosity: int = 0,
        verbosity_indent: str = "",
        **kwargs
) -> pd.DataFrame:
    """
    Compute motif score based on data set to obtain a matrix cell x TFBS motif.

    :param KMerClass kmer_obj: KMer object 
    :param KMerClass motif_kmer: Motif kmer histogram as KMerClass object. Should be created using the 
        function `motifs_to_kmer`
    :param str | None use_kmer_dr: Use a particular dimensionality reduced KMer representation saved under
        `kmer_obj.dr[use_kmer_dr]`.
    :param str | None use_motif_dr: Use a particular dimensionality reduced TFBS representation saved under
        `motif_kmer.dr[use_motif_dr]`.
    :param Callable | str | None norm: Normalize TFBS data using a spyce normalization function.
    :param pd.Series | pl.Series | np.ndarray | List | None norm_vec: Normalization vector. The function iterates over the unique values,
        creates a mask for them, and applies the `norm` function to each sub-matrix independently. If `None` is passed, no normalization is
        performed.
    :param pd.Series | pl.Series | np.ndarray | List | None harmony_vec: If set, perform Harmony batch correction over these value.
        Disable by setting this value to `None`.
    :param str | Callable collapse_fun: Define how the TFBS motif KMer probabilities are combined with
        the data after the k-mer histograms were multiplied. This obtains a TFBS-specific kmer 
        probability score per cell (a surrogate for the probability that the TFBS motif is present with a given
        k-mer in the data). `collapse_fun` defines how these k-mer specific TBFS probability scores are then aggregated to provide single
        score per TFBS motif. You can pass either a string (choose between `mean` | `max` | `min`) or a Callable.
        Note that the callable must aggregate over the kmer axis, which is `axis=1` per default.
    :param int verbosity: Verbosity level
    :param str verbosity_indent: Indentation string added before printout
    :params kwargs: Other keyword parameters passed to Harmony.
    :return pd.DataFrame: pandas dataframe containing the TF scoring in a matrix cells x motifs
    """
    tf_kmer_list = []
    if use_motif_dr is None:
        if not isinstance(motif_kmer.kmer_hist, np.ndarray):
            motif_data = sp.coo_matrix(motif_kmer.kmer_hist) 
            multiply_fun = sp.coo_matrix.multiply
        else:
            motif_data = motif_kmer.kmer_hist
            multiply_fun = np.multiply
    else:
        motif_data = motif_kmer.dr[use_motif_dr]
        multiply_fun = np.multiply

    if use_kmer_dr is None:
        iterator = tqdm(
            kmer_obj, 
            desc="%sCalculate motif score for data set" % verbosity_indent
        ) if verbosity > 0 else kmer_obj
    else:
        iterator = tqdm(
            kmer_obj.dr[use_kmer_dr], 
            desc="%sCalculate motif score for data set" % verbosity_indent
        ) if verbosity > 0 else kmer_obj.dr[use_kmer_dr]
    norm = convert_normalization(norm)
    
    # simple matrix operation
    # use row wise iteration for printout 
    for kmer in iterator:   
        expand_mat = multiply_fun(motif_data, kmer.reshape(1, -1))
        if isinstance(collapse_fun, Callable):
             tf_kmer_list.append(collapse_fun(expand_mat))
        else:
            if collapse_fun.lower() == "mean" or collapse_fun.lower() == "sum":  
                summed_val = expand_mat.sum(axis=1).reshape(-1, 1)
                if collapse_fun.lower() == "mean":
                    div = (expand_mat.sum(axis=0) > 0).sum()
                    if div == 0:
                        div = 1.
                    tf_kmer_list.append(summed_val / div)
                else:
                    tf_kmer_list.append(summed_val)
            elif collapse_fun.lower() == "max":
                tf_kmer_list.append(expand_mat.max(axis=1).reshape(-1, 1))
            elif collapse_fun.lower() == "min":
                tf_kmer_list.append(expand_mat.min(axis=1).reshape(-1, 1))
            else:
                raise NotImplementedError("The passed aggregation function is not implemented. Please pass your own "
                                          "callable (remember to aggregate over the kmer axis, axis=1 per default); or "
                                          "pass one of the following strings: `mean` | `max` | `min`.")

    stack_lib = np if isinstance(tf_kmer_list[0], np.ndarray) else sp
    data_mat = stack_lib.hstack(tf_kmer_list).T
    index = kmer_obj.barcodes
    cols = motif_kmer.barcodes
    if (data_mat.sum(axis=1) == 0).any():
        warnings.warn("Found cells with zero enrichment for all TFBS motifs. "
                      "Those cells are removed to avoid problems with further downstream correction procedures.")
        norm_vec = norm_vec[data_mat.sum(axis=1) > 0]
        harmony_vec = harmony_vec[data_mat.sum(axis=1) > 0]
        index = index[data_mat.sum(axis=1) > 0]
        data_mat = data_mat[data_mat.sum(axis=1) > 0]

    if (data_mat.sum(axis=0) == 0).any():
        warnings.warn("Found TFBS motifs that aren't present in any cell. " 
                      "Those TFBS motifs are removed to avoid problems with further downstream correction procedures.")
        cols = cols[data_mat.sum(axis=0) > 0]
        data_mat = data_mat[:, data_mat.sum(axis=0) > 0]

    if norm_vec is not None:
        if isinstance(norm_vec, pl.Series) or isinstance(norm_vec, pd.Series):
            norm_vec = norm_vec.to_numpy()
        elif isinstance(norm_vec, list):
            norm_vec = np.array(norm_vec)
        elif isinstance(norm_vec, np.ndarray):
            pass
        else:
            raise ValueError("`norm_vec` must be of type pl.Series | pd.Series | np.ndarray | List. "
                             "You passed a `norm_vec` of type %s" % type(norm_vec))
        if not isinstance(data_mat, np.ndarray):
            data_mat = data_mat.tocsr()

        for nval in np.unique(norm_vec):
            norm_mask = nval == norm_vec
            data_mat[norm_mask] = norm(data_mat[norm_mask], **kwargs)
    
    if harmony_vec is not None:
        data_mat = harmony_correct(data_mat, target_vec=harmony_vec, verbosity=verbosity - 1, **kwargs)
    
    # create dataframe
    cell_tf_mat = pd.DataFrame(
        data_mat if isinstance(data_mat, np.ndarray) else data_mat.A,
        index=index,
        columns=cols
    )
    return cell_tf_mat


def sig_test_beta(
        frg_sample: np.ndarray,
        bkg_sample: np.ndarray | float,
        motif_vec: pl.Series | pd.Series | np.ndarray | List,
        alternative: str = "greater",
        fdr_control: str = "bh",
        verbosity: int = 0,
        verbosity_indent: str = ""
) -> pl.DataFrame:
    """
    Compare TFBS scores of a foreground to a single background value (such as a score predicted over the all accessible regions/
    the entire genome) using a beta distribution. More precisely, fit TFBS scores to beta distribution and evaluate whether a 
    single (background) value is outside a confidence interval. 
    Note that this requires the scores to be properly normalised to yield a reasonable distribution.
    This method is experimental and should only be taken with a pinch of salt.

    :param np.ndarray frg_sample: Foreground data to which the beta distribution is approximated
    :param np.ndarray | float bkg_sample: Background value. This value is tested what the probability is that
        the value is part of the foreground data
    :param pl.Series | pd.Series | np.ndarray | List motif_vec: List with TFBS motif names
    :param str alternative: Test alternative. Choose between `greater` | `less` | `two-sided`.
    :param str fdr_control: Define algorithm for false discover rate control. Choose between `bh` | `by` | `none`
    :param int verbosity: Verbosity level
    :param str verbosity_indent: Indentation string added before printout
    :return pl.DataFrame: Return polars data frame with enrichment, p-value, and fdr per TFBS.
    """
    if isinstance(bkg_sample, float):
        bkg_sample = np.array([bkg_sample])
        frg_sample = frg_sample.reshape(-1, 1)
    if bkg_sample.shape[0] != frg_sample.shape[1] != motif_vec.shape[0]:
        raise ValueError("The number of background samples must match the number of columns in the foreground sample "
                         "and the number of motifs in `motif_vec`")
    sample_iter = zip(frg_sample.T, bkg_sample)
    if verbosity > 0:
        sample_iter = tqdm(
            sample_iter, 
            total=bkg_sample.shape[0],
            desc="%sCalulate beta significance" % verbosity_indent
        )
    pval_l, diff_l = [], []
    for frgs, bkgv in sample_iter:
        beta_params = st.beta.fit(frgs)
        diff_l.append(np.median(frgs) - bkgv)
        survival = st.beta(*beta_params).sf(bkgv)
        if alternative.lower() == "greater":
            pval = np.minimum(2 * (1. - survival), 1.)
        elif alternative.lower() == "lower" or alternative.lower() == "less":
            pval = np.minimum(2 * survival, 1.)
        elif alternative.lower() == "two-sided":
            pval = 2 * np.minimum(survival, 1. - survival)
        else:
            raise ValueError("Alternative not understood. Pass one of the following: `greater` | `less` | `two-sided`")
        pval_l.append(pval)

    pval_fdr = np.full_like(pval_l, fill_value=np.nan)
    if fdr_control is not None:
        if fdr_control.lower() != "none":
            if verbosity > 1:
                print("Apply %s FDR control" % fdr_control)
            pval_fdr = st.false_discovery_control(
                pval_l,
                method=fdr_control
            )
            
    return pl.DataFrame({
        "name": motif_vec,
        "absolute diff": diff_l,
        "p-value": pval_l,
        "fdr": pval_fdr
    })

def sig_tscore_test(
        frg_sample: np.ndarray,
        bkg_sample: np.ndarray | float,
        motif_vec: pl.Series | pd.Series | np.ndarray | List,
        alternative: str = "greater",
        fdr_control: str = "bh",
        verbosity: int = 0,
        verbosity_indent: str = ""
):
    """
    Compare TFBS scores of a foreground to a single background value (such as a score predicted over the all accessible regions/
    the entire genome) using a Student-t distribution. More precisely, we evaluate the difference between TFBS scores and
    a single (background), which should be t-distributed when the TFBS distribution is sufficiently close to normal. 
    Note that this requires the scores to be properly normalised to yield a reasonable distribution.
    This method is experimental and should only be taken with a pinch of salt.

    :param np.ndarray frg_sample: Foreground data to which we measure the difference to the backround. Data is expected to
        be close to normally distributed.
    :param np.ndarray | float bkg_sample: Background value. It is tested whether values in the forground sample are significantly
        different from the mean.
    :param pl.Series | pd.Series | np.ndarray | List motif_vec: List with TFBS motif names
    :param str alternative: Test alternative. Choose between `greater` | `less` | `two-sided`.
    :param str fdr_control: Define algorithm for false discover rate control. Choose between `bh` | `by` | `none`
    :param int verbosity: Verbosity level
    :param str verbosity_indent: Indentation string added before printout
    :return pl.DataFrame: Return polars data frame with enrichment, p-value, and fdr per TFBS.
    """
    if isinstance(bkg_sample, float):
        bkg_sample = np.array([bkg_sample])
        frg_sample = frg_sample.reshape(-1, 1)
    if bkg_sample.shape[0] != frg_sample.shape[1] != motif_vec.shape[0]:
        raise ValueError("The number of background samples must match the number of columns in the foreground sample "
                         "and the number of motifs in `motif_vec`")
    sample_iter = zip(frg_sample.T, bkg_sample)
    if verbosity > 0:
        sample_iter = tqdm(
            sample_iter, 
            total=bkg_sample.shape[0],
            desc="%sCalulate beta significance" % verbosity_indent
        )
    pval_l, diff_l = [], []
    for frgs, bkgv in sample_iter:
        frg_mean = frgs.mean()
        frg_std = frgs.std()
        diff_l.append(frg_mean - bkgv)
        t_x = (bkgv - frg_mean) / frg_std
        survival = st.t(df=frgs.shape[0] - 1).sf(t_x) 
        if alternative.lower() == "greater":
            pval = np.minimum(2 * (1. - survival), 1.)
        elif alternative.lower() == "lower" or alternative.lower() == "less":
            pval = np.minimum(2 * survival, 1.)
        elif alternative.lower() == "two-sided":
            pval = 2 * np.minimum(survival, 1. - survival)
        else:
            raise ValueError("Alternative not understood. Pass one of the following: `greater` | `less` | `two-sided`")
        pval_l.append(pval)

    pval_fdr = np.full_like(pval_l, fill_value=np.nan)
    if fdr_control is not None:
        if fdr_control.lower() != "none":
            if verbosity > 1:
                print("Apply %s FDR control" % fdr_control)
            pval_fdr = st.false_discovery_control(
                pval_l,
                method=fdr_control
            )
            
    return pl.DataFrame({
        "name": motif_vec,
        "absolute diff": diff_l,
        "p-value": pval_l,
        "fdr": pval_fdr
    })

def sig_test_over_total_background(
       kmer_obj: KMerClass,
       motif_kmer: KMerMatrix,
       total_bkg_kmer: KMerClass,
       ct_vec: pd.Series | pl.Series | np.ndarray | List | None = None,
       norm: Callable | str | None = "sum_norm",
       norm_vec: pd.Series | pl.Series | np.ndarray | List | None = None,
       harmony_vec: pd.Series | pl.Series | np.ndarray | List | None = None,
       collapse_fun: str | Callable = "sum",
       alternative: str = "greater",
       sig_test: str = "t",
       fdr_control: str | None = "bh",
       verbosity: int = 0,
       verbosity_indent: str = "",
       **kwargs
) -> Dict[str, pl.DataFrame] | pl.DataFrame:
    """
    Perform signficance test using a non-normalized `kmer_obj` and the k-mer histogram over all 
    features and peaks (background). Significance is tested by calculating a beta distribution over
    the foreground values (ie. normalized probabilities mulitplied by a constant) and the probability of the 
    background value. Please pass a non-normalized `kmer_obj` and normalize using `sum_norm`. This only valid
    when using a single-species.
    This method is experimental and should only be taken with a pinch of salt.

    :param KMerClass kmer_obj: KMer data object
    :param KMerMatrix motif_kmer: KMer matrix of motifs.
    :param KMerClass total_bkg_kmer: KMer object representing k-mer histogram/scores over all accessible regions.
    :param pd.Series | pl.Series | np.ndarray | List | None ct_vec: Vector with cell type labels. Size matches
        number of cells in KMer object
    :param Callable | str | None norm: Normalization applied to TFBS scores
    :param pd.Series | pl.Series | np.ndarray | List | None norm_vec: Normalization vector indicating groups of cells
        that are normalized together (for example for centering).
    :param pd.Series | pl.Series | np.ndarray | List | None harmony_vec: Vector with cell labels that are used
        for the harmony correction.
    :param str | Callable collapse_fun: Aggregation function used for collapsing the TFBS scores over k-mers
    :param str alternative: Alternative for significance test. Choose between `greater` | `less` | `two-sided`
    :param str sig_test: Significance test to perform. Choose between `t` | `beta`.
    :param str fdr_control: Type of control applied to account for false discovery rate.
    :param int verbosity: Verbosity level
    :param str verbosity_indent: Indentation string added before printout
    :param kwargs: Other keywords that are passed the motif score function
    :return Dict[str, pl.DataFrame] | pl.DataFrame: If `ct_vec` is passed, return an enrichment table per cell type, 
        otherwise return results directly (containing name, absolute difference, p-value and q-value for TFBS).
    """
    
    
    if norm_vec is None:
        norm_vec = np.ones(kmer_obj.shape[0])

    if sig_test.lower() == "t":
        sig_fun = sig_tscore_test
    elif sig_test.lower() == "beta":
        sig_fun = sig_test_beta
    else:
        raise ValueError("`sig_test` not understood. Please pass one of the following: `t` | `beta`.")
    
    tfbs_score_df = data_set_motif_score(
        kmer_obj=kmer_obj,
        motif_kmer=motif_kmer,
        norm=norm,
        norm_vec=norm_vec,
        harmony_vec=harmony_vec,
        collapse_fun=collapse_fun,
        verbosity=verbosity,
        verbosity_indent=verbosity_indent,
        **kwargs
    )

    total_tfbs_score_df = data_set_motif_score(
        kmer_obj=total_bkg_kmer,
        motif_kmer=motif_kmer,
        norm=norm,
        norm_vec=np.ones(1),
        harmony_vec=harmony_vec,
        collapse_fun=collapse_fun,
        verbosity=verbosity - 1,
        verbosity_indent=verbosity_indent,
        **kwargs
    )
    
    if ct_vec is None:
        return sig_fun(
            tfbs_score_df.to_numpy(),
            total_tfbs_score_df.to_numpy().flatten(),
            motif_vec=total_tfbs_score_df.columns,
            alternative=alternative,
            fdr_control=fdr_control,
            verbosity=verbosity,
            verbosity_indent=verbosity_indent
        )
    else:
        sig_test_results = {}
        for ct in ct_vec.unique():
            ct_mask = (ct == ct_vec).to_numpy()
            sig_test_results[ct] = sig_fun(
                tfbs_score_df[ct_mask].to_numpy(),
                total_tfbs_score_df.to_numpy().flatten(),
                motif_vec=total_tfbs_score_df.columns,
                alternative=alternative,
                fdr_control=fdr_control,
                verbosity=verbosity,
                verbosity_indent=verbosity_indent + ct + ": "
            )
        return sig_test_results


def sig_test(
        frg_sample: np.ndarray,
        bkg_sample: np.ndarray,
        motif_vec: pl.Series | pd.Series | np.ndarray | List,
        test_type: str = "mannwhitneyu",
        alternative: str = "greater",
        fdr_control: str | None = "bh",
        verbosity: int = 0,
        verbosity_indent: str = "",
        **kwargs
) -> pl.DataFrame:
    """
    Perform significance test for difference over given samples.

    :param np.ndarray frg_sample: foreground TFBS kmer sample (cells x feature)
    :param np.ndarray bkg_sample: background TFBS kmer sample (cells x feature)
    :param pd.Series | pl.Series | np.ndarray | List | None motif_vec: Vector with motif names.
    :param str test_type: Type of significance test. Choose between `mannwhitneyu` | `ks` for Kolmogorov-Smirnov
    :param str alternative: Hypotehsis test. Pass `greater` | `less` | `two-sided`
    :param str | None fdr_control: False-discovery control. Pass `bh` for Benjamini-Hochberg | `by` for
        Benjamini-Yekutieli | `none` for none.
    :param int verbosity: Verbosity level
    :param str verbosity_indent: Prefix for progress bar and print outs
    :param kwargs: Unused parameters.
    :return pl.Series: pvalues, fdr control (if set), statistics, and difference per motif.
    """
    if test_type.lower() == "ks":
        test_fun = st.ks_2samp
    elif test_type.lower() == "mannwhitneyu":
        test_fun = st.mannwhitneyu
    else:
        raise ValueError("Test statistic %s not understood. Please pass one of the following: ks | mannwhitneyu." % test_type)
    pval  = np.zeros(frg_sample.shape[1])
    stats = np.zeros(frg_sample.shape[1])
    data_iter = tqdm(
        np.arange(frg_sample.shape[1]),
        total=frg_sample.shape[1],
        desc="%sSignificant test" % verbosity_indent
    ) if verbosity > 0 else np.arange(frg_sample.shape[1])
    for data_idx in data_iter:
        res = test_fun(
            frg_sample[:, data_idx], 
            bkg_sample[:, data_idx], 
            alternative=alternative
        )
        pval[data_idx] = res.pvalue
        stats[data_idx] = res.statistic
        
    frg_mean = np.median(frg_sample, axis=0)
    bkg_mean = np.median(bkg_sample, axis=0)   
    div = np.std(frg_sample, axis=0)
    div[div == 0] = 1.
    diff = (frg_mean - bkg_mean) / div

    pval_fdr = np.full_like(pval, fill_value=np.nan)
    if fdr_control is not None:
        if fdr_control.lower() != "none":
            if verbosity > 1:
                print("Apply %s FDR control" % fdr_control)
            pval_fdr = st.false_discovery_control(
                pval,
                method=fdr_control
            )
            
    return pl.DataFrame({
        "name": motif_vec,
        "absolute diff": diff,
        "statistic": stats,
        "p-value": pval,
        "fdr": pval_fdr
    })


def equivalence_test(
        frg_sample: np.ndarray,
        bkg_sample: np.ndarray,
        motif_vec: pl.Series | pd.Series | np.ndarray | List,
        fdr_control: str | None = "bh",
        scale_dstat: bool = True,
        verbosity: int = 0,
        verbosity_indent: str = "",
        **kwargs
):
    r"""
    **Note that this method is experimental**. Equivalence test for Kolmogornov-Smirnov test taken from

    > Alexis (https://stats.stackexchange.com/users/44269/alexis), 
    > Is there a simple equivalence test version of the Kolmogorov–Smirnov test?, 
    > URL (version: 2021-04-14): https://stats.stackexchange.com/q/108487

    and adapted according to

    > Hodges Jr, J. L. "The significance probability of the Smirnov two-sample test." 
    > Arkiv för matematik 3.5 (1958): 469-486.

    Note that, despite the fact that there's a parameter for foreground and background values, the
    significance test is symmetric. Swapping foreground and background will provide the same results.
    Note that as pointed out in the Stackexchange comment, a p-value = 0.05 might not be the best
    approach, as the underlying Rayleigh distribution is non-symmetric which increases quickly for low 
    values with a sigma of 0.5 (which is the case for this equivalence test).
    The critical value is computed as in 

    .. math::
        Q(p) = \sqrt{\frac{-\ln{(1 - p)}}{2}}
    
    
    Given the variance the Rayleigh distribution is 0.5, the comment suggests a conservative threshold of
    
    .. math::
        Q(0.05) + \frac{\sigma}{4} = Q(0.05) + \frac{1}{8}
    
    
    (which is around 0.15); or a more liberal threshold when adding half of the variance, which results in a
    p-value of around 0.28.

    :param np.ndarray frg_sample: Foreground data values (KMer values or TFBS scores).
    :param np.ndarray bkg_sample: Background data values (KMer values or TFBS scores). 
    :param pl.Series | pd.Series | np.ndarray | List motif_vec: List or iterable with motif names
    :param str | None fdr_control: Algorithm for False-Discovery-Rate (FDR) control. Pass `bh` 
        for Benjamini-Hochberg | `by` for Benjamini-Yekutieli | `none` for none.
    :param bool scale_dstat: If `True`, scale KS statistic according to the number of data points used. 
    :param int verbosity: Verbosity level
    :param str verbosity_indent: Prefix for progress bar and print outs
    :param kwargs: Unused parameters.
    :return pl.Series: pvalues, fdr control (if set), statistics, and difference per motif.
    """

    pval  = np.zeros(frg_sample.shape[1])
    stats = np.zeros(frg_sample.shape[1])
    data_iter = tqdm(
        np.arange(frg_sample.shape[1]),
        total=frg_sample.shape[1],
        desc="%sEquivalence test" % verbosity_indent
    ) if verbosity > 0 else np.arange(frg_sample.shape[1])

    n_frg = frg_sample.shape[0]
    n_bkg = bkg_sample.shape[0]
    sample_scale = np.sqrt(n_frg * n_bkg / (n_frg + n_bkg))
    for data_idx in data_iter:
        # calculate D (maximum of D+ and D-)
        D_stat = st.ks_2samp(
            frg_sample[:, data_idx], 
            bkg_sample[:, data_idx], 
            alternative="two-sided"
        ).statistic  

        # scale D according to sample size
        D_sample = D_stat * sample_scale

        # use result in Hodges 1958 and stackexchange comment to compute p-value
        pval[data_idx] = st.rayleigh.cdf(D_sample if scale_dstat else D_stat, scale=0.5)
        stats[data_idx] = D_stat

    frg_mean = np.median(frg_sample, axis=0)
    bkg_mean = np.median(bkg_sample, axis=0)   
    diff = frg_mean - bkg_mean

    pval_fdr = np.full_like(pval, fill_value=np.nan)
    if fdr_control is not None:
        if fdr_control.lower() != "none":
            if verbosity > 1:
                print("Apply %s FDR control" % fdr_control)
            pval_fdr = st.false_discovery_control(
                pval,
                method=fdr_control
            )
            
    return pl.DataFrame({
        "name": motif_vec,
        "absolute diff": diff,
        "statistic": stats[data_idx],
        "p-value": pval,
        "fdr": pval_fdr
    })
   


def find_sig_tf(
        ct_vec: pd.Series | pl.Series | np.ndarray | List,
        kmer_obj: KMerClass | None = None,
        motif_kmer: KMerClass | None = None,
        tfbs_score: pd.DataFrame | None = None,
        norm: None | str | Callable = None,
        norm_vec: pd.Series | pl.Series | np.ndarray | List | None = None,
        harmony_vec: pd.Series | pl.Series | np.ndarray | List | None = None,
        use_kmer_dr: str | None = None,
        use_motif_dr: str | None = None,
        collapse_fun: str | Callable = "sum",
        sig_test_type: str = "divergence",
        tfbs_test_type: str = "ks",
        tfbs_alternative: str = "greater",
        tfbs_fdr_control: str | None = "bh",
        scale_by_size: bool = True,
        avg_policy: str = "median",
        verbosity: int = 0,
        verbosity_indent: str = "", 
        **kwargs
) -> Dict[str, pl.DataFrame]:
    """
    TFBS motif enrichment signficance or equivalence test. The function requires a cell type annotation and
    either the pre-computed TFBS scores or a KMer object for the data set and for the
    TFBS motifs (see function `motifs_to_kmer`). It then iterates over all cell types and comparse the TFBS
    distribution with all other cell types. Note that this circumvents the problem of a significance test
    over a multimodal distribution. Finally, all tests are aggregated together to yield a single
    enrichment, p, and q-value per cell type.

    :param pd.Series | pl.Series | np.ndarray | List ct_vec: Cell type annotation or cell grouping as a list, array
        or series.
    :param KMerClass | None kmer_obj: `KMerMatrix` or `KMerCollection` that contains the data KMer data. Only necessary when
        TFBS scores haven't been pre-computed.
    :param KMerClass | None motif_kmer: `KMerMatrix` for the TFBS motifs. The matrix can be created using the function
        `motifs_to_kmer`. Only necessary if TFBS scores haven't been pre-computed.
    :param pd.DataFrame | None tfbs_score: Pre-computed TFBS scores. 
    :param  None | str | Callable norm: Normalize with spyce normalization faction along `norm_vec`. To be more precise,
        The function iterates over unique values in `norm_vec`, creates a mask, and applies `norm` to the masked sub-matrices
        independently. 
    :param pd.Series | pl.Series | np.ndarray | List | None norm_vec: Normalization vector. The function iterates over the unique values,
        creates a mask for them, and applies the `norm` function to each sub-matrix independently. If `None` is passed, no normalization is
        performed.
    :param pd.Series | pl.Series | np.ndarray | List | None harmony_vec: If set, perform Harmony batch correction over these value.
        Disable by setting this value to `None`.
    :param str | None use_kmer_dr: Use a specific k-mer dimensionality reduction for creating the TFBS enrichment. If `None`,
        use the k-mer data directly.
    :param str | None use_motif_dr: Use a specific motif dimensionality reduction for creating the TFBS enrichment. If `None`,
        use the k-mer data directly.
    :param str | Callable collapse_fun: Aggregation function used for creating the TFBS motif enrichemnt score
        (for more information, see function `data_set_motif_score`).
    :param str sig_test_type: Pass `conservation` if you want to test for significant similarity, `divergence` if you want to 
        test for significant difference. Note that if you pass `conservation`, the `tfbs_test_type` will be automatically 
        set to `ks`. **Note that `conservation` is experimental.**
    :param str tfbs_test_type: Test type for significant TFBS motifs. Choose between `mannwhitneyu` and `ks`. Note that
        when choosing `sig_test_type=conservation`, this will be automatically set to `ks`. If you've set
        `mannwhitneyu` nonetheless, the function will raise a warning.
    :param str tfbs_alternative: Alternative for TFBS significant test. Choose between `greater` | `less` | `two-sided`.
        This parameter is ignored for `sig_test_type=conservation`
    :param str | None tfbs_fdr_control: FDR correction algorithm for TFBS signficant test. Pass `bh` for Benjamini-Hochberg 
        | `by` for Benjamini-Yekutieli | `none` for none.
    :param bool scale_by_size: If `True`, weigh results according to number of cells per type before aggregating using 
        `avg_policy` (does not apply when `avg_policy="fdr"`).
    :param str avg_policy: Significance tests are performed between every pair of cell types. `avg_policy` defines how 
        these results are combined to a single significance value for each cell type and TFBS. Choose between
        `mean` | `max_fdr` | `median` | `min_p` | `max_diff`.
    :param int verbosity: Verbosity levels
    :param str verbosity_indent: Prefix for output and progress bars.
    :param kwargs: Other key word parameters that are passed to subfunctions.
    :return Dict[str, pl.DataFrame]: A dictionary with the group name as key and the data frame with TFBS motif 
        enrichement results as value.
    """
    if isinstance(ct_vec, list) or isinstance(ct_vec, np.ndarray):
        ct_vec = pd.Series(ct_vec)

    if norm_vec is None:
        norm_vec = ct_vec

    if sig_test_type.lower() == "conservation" and tfbs_test_type.lower() == "mannwhitneyu":
        warnings.warn("You've passed `sig_test_type='conservation'` and `tfbs_test_type='mannwhitneyu'`. Significance test for "
                      "conservation are currently only provided for `ks` tests. `tfbs_test_type='mannwhitneyu'` will be ignored.")

    if sig_test_type.lower() == "divergence":
        compare_fun = sig_test 
    elif sig_test_type.lower() == "conservation":
        compare_fun = equivalence_test
    else:
        raise ValueError("`sig_test_type=%s` not understood. Please pass one of the following: `divergence` | `conservation`." % sig_test_type)
    
    if tfbs_score is None and kmer_obj is not None and motif_kmer is not None:
        if isinstance(harmony_vec, pl.Series):
            harmony_vec = harmony_vec.to_pandas()
        if (pd.Series(harmony_vec).value_counts() < 2).any():
            raise ValueError("There must be at least 2 entries per group in `harmony_vec`.")
        tfbs_score = data_set_motif_score(
            kmer_obj=kmer_obj,
            motif_kmer=motif_kmer,
            use_kmer_dr=use_kmer_dr,
            use_motif_dr=use_motif_dr,
            norm=norm,
            norm_vec=norm_vec,
            harmony_vec=harmony_vec,
            verbosity=verbosity,
            verbosity_indent=verbosity_indent,
            collapse_fun=collapse_fun,
            **kwargs
        )
    elif tfbs_score is not None:
        pass
    else:
        raise ValueError("You need to pass either a pre-computed TFBS score matrix or a data AND a motif KMerMatrix/KMerCollection.")
    
    n_cells = ct_vec.value_counts()
    ct_set = ct_vec.unique()
    ct_iterator = tqdm(
        ct_set,
        desc="%sTFBS for cell type" % verbosity_indent
    ) if verbosity > 0 else ct_set
    
    ct_sig_tfbs_dict = {}
    for ct in ct_iterator:
        other_ct = n_cells.drop(index=ct)
        ct_mask = (ct_vec == ct).to_numpy().astype("bool")
        if ct_mask.astype("int").sum() <= 5:
            warnings.warn("You must have more than 5 cells per annotated foreground cell type. Skip %s." % ct)
            continue

        # fetch tfbs scoring for foreground
        frg_samples = tfbs_score.iloc[ct_mask].to_numpy()
        ct_df_l = []
        for bkg_ct in ct_set:
            if bkg_ct == ct: continue
            bkg_mask = (ct_vec == bkg_ct).to_numpy().astype("bool")
            if bkg_mask.astype("int").sum() <= 5:
                warnings.warn("You must have more than 5 cells per annotated background cell type. Skip %s." % bkg_ct)
                continue
            
            # fetch tfbs scoring forbackground
            bkg_samples = tfbs_score.iloc[bkg_mask].to_numpy()

            if frg_samples.shape[0] == 0 or bkg_samples.shape[0] == 0: 
                continue
            
            bct_df = compare_fun(
                frg_sample=frg_samples,
                bkg_sample=bkg_samples,
                motif_vec=tfbs_score.columns,
                test_type=tfbs_test_type,
                alternative=tfbs_alternative,
                verbosity=verbosity - 1,
                verbosity_indent="%s\t" % verbosity_indent,
                fdr_control=tfbs_fdr_control,
                **kwargs
            )
    
            ct_df_l.append(
                bct_df.group_by("name").agg(pl.col(["absolute diff", "statistic", "p-value", "fdr"]).median()).with_columns(pl.lit(bkg_ct).alias("background"))
            )
            if scale_by_size and avg_policy.lower() != "fdr":
                scaling = other_ct.loc[bkg_ct] / float(other_ct.sum())
                ct_df_l[-1] = ct_df_l[-1].with_columns(
                    pl.col(["absolute diff", "statistic", "p-value", "fdr"]) * scaling
                )
        if len(ct_df_l) > 0:
            ct_sig_tfbs_dict[ct] = pl.concat(ct_df_l)
        else:
            ct_sig_tfbs_dict[ct] = pl.DataFrame(schema=["name", "absolute diff", "statistic", "p-value", "fdr"])
            continue
        if ct_sig_tfbs_dict[ct].shape[0] > 0:
            # significance is set to the largest fdr
            # determine average difference
            if avg_policy.lower() == "mean":
                ct_sig_tfbs_dict[ct] = ct_sig_tfbs_dict[ct].group_by("name").agg(pl.mean([ "absolute diff", "statistic", "p-value", "fdr"]))
            elif avg_policy.lower() == "max_fdr":
                ct_sig_tfbs_dict[ct] = ct_sig_tfbs_dict[ct].group_by("name").agg(
                    pl.all().sort_by(["fdr", "p-value"], descending=True).first()
                ).drop("background")
            elif avg_policy.lower() == "min_p":
                ct_sig_tfbs_dict[ct] = ct_sig_tfbs_dict[ct].group_by("name").agg(
                    pl.all().sort_by(["p-value"], descending=False).first()
                ).drop("background")
            elif avg_policy.lower() == "max_diff":
                ct_sig_tfbs_dict[ct] = ct_sig_tfbs_dict[ct].group_by("name").agg(
                    pl.all().sort_by(["absolute diff"], descending=True).first()
                ).drop("background")
            elif avg_policy.lower() == "median":
                ct_sig_tfbs_dict[ct] = ct_sig_tfbs_dict[ct].group_by("name").agg(pl.median(["absolute diff",  "statistic", "p-value", "fdr"]))
            else:
                raise ValueError("`avg_policy` not understood. Please pass one of the following: `mean` | `max_fdr` | `median` | `min_p` | `max_diff`.")
    return ct_sig_tfbs_dict    

